import numpy as np
import os
from pathlib import Path
import pickle
import datetime as dt
from coffee_roulette import utils


def create_coffee_roulette(roulette_name, working_dir=None, participants=None):
    """
    Creates a new coffee roulette project directory.

    This involves files for storing a meeting probability matrix
    and a dictionary of participant names.

    This function only needs to be run once.

    PARAMETERS
    -------------
        roulette_name : string
            The name of the roulette instance

        working_dir : string, optional
            Path to the desired file directory.
            Defaults to the current working directory.

        participants : list, option
            Lists of participants.
            If None, the user is prompted to enter names.

    WRITES
    -------------
    working_dir/crParticipantDict.pkl
    working_dir/crProbMatrix.npy
    working_dir/crConversationStarters.txt

    """
    rdate = dt.datetime.today().strftime("%Y-%m-%d")

    if working_dir == None:
        working_dir = "."
    wd = Path(working_dir).resolve()

    roulette_dir = wd / f"CoffeeRoulette-{roulette_name}"

    if os.path.exists(roulette_dir):
        raise ValueError(
            "Coffee roulette already exists! Run another function or choose a different roulette name!"
        )

    os.mkdir(roulette_dir)
    os.mkdir(Path(roulette_dir) / "results")

    # create participant dictionary (names and who they have met)
    crParticipantDict = {}
    if participants == None:
        id = 0
        while True:
            participant = input(
                "Enter participant name or 'x' if you are done for now: "
            )
            if participant == "x":
                break
            crParticipantDict[id] = (participant, [])
            id += 1
    elif type(participants) == list:
        for id, participant in enumerate(participants):
            crParticipantDict[id] = (participant, [])
    else:
        raise ValueError("Participants must be a flat list or None!")

    pickle.dump(
        crParticipantDict, open(roulette_dir / "crParticipantDict.p", "wb")
    )

    # create pair probability matrix
    crProbMatrix = np.full((id, id), 1 / (id - 1))
    np.fill_diagonal(crProbMatrix, 0)
    np.save(roulette_dir / "crProbMatrix.npy", crProbMatrix)

    # create a config file
    cfg_file, ruamelFile = utils.create_config_template()
    cfg_file["roulette_name"] = roulette_name
    cfg_file["roulette_date"] = rdate
    cfg_file["roulette_directory"] = str(roulette_dir)
    cfg_file["roulette_frequency"] = 3  # weeks
    cfg_file["roulette_times"] = [[9, 12], [13, 17]]  # hours
    cfg_file["time_spacing"] = 30  # minutes
    cfg_file[
        "name_prefix"
    ] = "@"  # for places with usernames (slack, discord, etc)
    cfg_file["meeting_purpose"] = "have coffee"

    cfg_path = roulette_dir / "config.yaml"
    utils.write_config(cfg_path, cfg_file)

    lines = [
        "What is the most bizarre thing that you have ever eaten?",
        "What is the worst advice that you have ever taken?",
        "What is the magical power that you wish you had?",
        "If you could trade lives with one person, who would it be?",
        "What is one thing you would tell yourself 10 years ago?",
        "What is your biggest pet peeve?",
        "What about our society would surprise a time traveller?",
        "What is a funny prank you have played on someone?",
        "What is a secret talent you have?",
        "What is one embarrassing memory you have?",
        "If you could have dinner with one person, dead or alive, who would it be?",
    ]
    with open(roulette_dir / "crConversationStarters.txt", "w") as f:
        f.writelines(line + "\n" for line in lines)

    print(
        "\nA new project with name 'CoffeeRoulette-{roulette_name}-{date}' is created at {roulette_dir}.\nUse function 'generate_pairs' to run the first round of the coffee roulette.\nIf you wish to customise the conversation starters, edit the crConversationStarters.txt file in your roulette directory.\nYou can add and remove participants at any time by running the functions 'add_participants' and 'remove_participants' respectively."
    )



from pathlib import Path
import numpy as np
import pickle
import os
import datetime as dt
import random
from coffee_roulette import utils


def get_pairs(
    cfg_path, conv_starters=True, random_time=True, exclude_nonrandom=None
):
    """
    Generates pairs of participants based on the probability matrix
    such that people can only be paired with those they have not been
    paired with before.

    Prints the freshly generated pairs and, in case of an
    odd number of participants, the name of the left-out person.

    PARAMETERS
    -------------
        cfg_path : string
            Full path to the config.yaml file
        conv_starters : bool, optional
            Whether to include suggested conversation starters in the output file
            Default is True
        random_time : bool, optional
            Whether to include a random meeting time within the next
            config["roulette_frequency"] weeks at times config["roulette_times"].
            Use edit_config() to edit the config file programmatically.
            Defaults to True
        exclude_nonrandom : int or string, optional
            Participant id or name to exclude
            Find out participant ids and names by running print_participants()
            Defaults to None, which excludes someone randomly
    """
    cfg = utils.read_config(cfg_path)
    roulette_dir = Path(cfg["roulette_directory"]).resolve()

    everyone_matched = False
    iteration = 1
    output_list = []
    while not everyone_matched:
        print(f"Running the iteration #{iteration}...")
        # load_files
        crProbMatrix = np.load(roulette_dir / "crProbMatrix.npy")
        crParticipantDict = pickle.load(
            open(roulette_dir / "crParticipantDict.p", "rb")
        )

        used_ids = np.empty(0)
        participant_number = crProbMatrix.shape[0]

        if os.path.exists(roulette_dir / "crRemovedParticipantDict.p"):
            crRemovedParticipantDict = pickle.load(
                open(roulette_dir / "crRemovedParticipantDict.p", "rb")
            )
            used_ids = np.append(
                used_ids, list(crRemovedParticipantDict.keys())
            )
            used_ids = used_ids.flatten()
            participant_number = participant_number - len(
                crRemovedParticipantDict
            )

        # if odd number of participants, have the participant id'ed by exclude_nonrandom sit the round out
        if participant_number % 2 == 1 and exclude_nonrandom != None:
            used_ids = np.append(used_ids, [exclude_nonrandom])
            participant_number = participant_number - len(exclude_nonrandom)

        # set up meeting timeframes
        if random_time:
            weeks = cfg["roulette_frequency"]
            times = cfg["roulette_times"]
            spacing = cfg["time_spacing"]

            t = dt.date.today()  # the day the code is run
            start_date = t + dt.timedelta(days=1)
            day_number = int(7 * weeks)  # days until the next draw

            reasonable_times = utils.get_meeting_times(times, spacing)

        for p in range(int(participant_number / 2)):
            unused_ids = np.setdiff1d(
                np.arange(crProbMatrix.shape[0]), used_ids
            )
            max_val_remaining_crProbMatrix = crProbMatrix[
                np.repeat(unused_ids, unused_ids.shape),
                np.tile(unused_ids, unused_ids.shape),
            ].max()
            i, j = np.where(crProbMatrix == max_val_remaining_crProbMatrix)
            for x in used_ids:
                ids_to_remove = np.where(i == x)
                i = np.delete(i, ids_to_remove)
                j = np.delete(j, ids_to_remove)
                jds_to_remove = np.where(j == x)
                i = np.delete(i, jds_to_remove)
                j = np.delete(j, jds_to_remove)
            id = np.random.choice(len(i))
            output_str = f"- @{crParticipantDict[i[id]][0]} will {cfg['meeting_purpose']} with @{crParticipantDict[j[id]][0]}"

            # generate a random meeting time
            if random_time:
                meeting_weekday = "Saturday"
                while (
                    meeting_weekday == "Saturday"
                    or meeting_weekday == "Sunday"
                ):
                    random_number_of_days = random.randrange(day_number)
                    meeting_date = start_date + dt.timedelta(
                        random_number_of_days
                    )
                    meeting_weekday = meeting_date.strftime("%A")
                meeting_time = random.choice(reasonable_times)
                output_str += f". Perhaps on {meeting_date.strftime('%A')}, {meeting_date.strftime('%B %d')}, at {':'.join(str(meeting_time).split(':')[:2])}?"

            # generate a conversation starter
            if conv_starters:
                # find conv starters not used by either participant
                used_conv_starters = (
                    crParticipantDict[i[id]][1] + crParticipantDict[j[id]][1]
                )
                all_conv_starters = np.arange(
                    len(
                        open(
                            roulette_dir / "crConversationStarters.txt"
                        ).readlines()
                    )
                )
                available_conv_starters = np.setdiff1d(
                    all_conv_starters, used_conv_starters
                )

                # pick a random question
                if available_conv_starters.shape[0] > 0:
                    q_id = np.random.choice(available_conv_starters)
                    with open(
                        roulette_dir / "crConversationStarters.txt", "r"
                    ) as text_file:
                        for l, line in enumerate(text_file):
                            if l == q_id:
                                question = line
                else:
                    print(
                        "All conversation starters have been used! Add more!"
                    )
                    break
                # add the question id to participants' conversation starter history
                crParticipantDict[i[id]][1].append(q_id)
                crParticipantDict[j[id]][1].append(q_id)
                output_str += f" Suggested conversation starter: {question} "

            # check that the last pairing has not met already
            if (
                crProbMatrix[i[id], j[id]] != 0
            ):  # all good - these people have not met yet!
                print(output_str)
                output_list.append(output_str)
                used_ids = np.concatenate((used_ids, [i[id], j[id]]))
                crProbMatrix[i[id], j[id]] = 0
                crProbMatrix[j[id], i[id]] = 0
                if p == (int(participant_number / 2)) - 1:
                    print(f"Success! Number of iterations needed: {iteration}")
                    everyone_matched = True
            else:  # problem! break out of the for loop, start over!
                iteration += 1
                break

    # print a line about an unpaired participant if the total number is odd and exclusion is random
    if participant_number % 2 == 1:
        unpaired_id = np.setdiff1d(np.arange(crProbMatrix.shape[0]), used_ids)
        unpaired_str = f"{crParticipantDict[unpaired_id[0]][0]} was not paired with anyone this time!"
        print(unpaired_str)
        output_list.append(unpaired_str)

    # recompute probabilities and update the matrix
    for row in range(crProbMatrix.shape[0]):
        nonzeros = crProbMatrix[row] != 0
        crProbMatrix[row, nonzeros] = 1 / nonzeros.sum()

    rdate = dt.datetime.today().strftime("%Y-%m-%d")
    rtime = dt.datetime.today().strftime("%H%M%S")

    with open(
        roulette_dir / "results" / f"{rdate}_crPairings_{rtime}.txt", "w"
    ) as f:
        f.writelines(line + "\n" for line in output_list)
    np.save(roulette_dir / "crProbMatrix.npy", crProbMatrix)
    pickle.dump(
        crParticipantDict, open(roulette_dir / "crParticipantDict.p", "wb")
    )



from coffee_roulette import utils

def test_get_meeting_times():
    import datetime as dt

    assert utils.get_meeting_times([[9, 10]], spacing=30) == [
        dt.timedelta(seconds=32400),
        dt.timedelta(seconds=34200),
    ]
    assert utils.get_meeting_times(
        [[900, 1000], [1430, 1630]], spacing=50
    ) == [
        dt.timedelta(seconds=32400),
        dt.timedelta(seconds=35400),
        dt.timedelta(seconds=52200),
        dt.timedelta(seconds=55200),
        dt.timedelta(seconds=58200),
    ]





import os
import numpy as np
from ruamel.yaml import YAML
from pathlib import Path


def create_config_template():
    """
    Creates a template for the config.yaml file.
    Code inspired by DeepLabCut https://github.com/DeepLabCut/DeepLabCut/blob/master/AUTHORS
    """
    yaml_str = """\
    # Roulette definitions
        roulette_name:
        roulette_date:
        roulette_directory:
        \n
    # Pairing parameters
        roulette_frequency:
        roulette_times:
        time_spacing:
        \n
    # Printing parameters
        name_prefix: 
        meeting_purpose:
        \n    
    """

    ruamelFile = YAML()
    cfg_file = ruamelFile.load(yaml_str)
    return cfg_file, ruamelFile


def read_config(cfg_path):
    """
    Reads the roulette config file
    """
    ruamelFile = YAML()
    cfg_path = Path(cfg_path)
    if os.path.exists(cfg_path):
        with open(cfg_path, "r") as f:
            cfg = ruamelFile.load(f)
    else:
        raise FileNotFoundError("Config file not found!")
    return cfg


def write_config(cfg_path, cfg):
    """
    Writes the roulette config file
    """
    with open(cfg_path, "w") as f:
        cfg_file, ruamelFile = create_config_template()
        for key in cfg.keys():
            cfg_file[key] = cfg[key]
        ruamelFile.dump(cfg_file, f)


def edit_config(cfg_path, edits):
    cfg = read_config(cfg_path)
    for key, value in edits.items():
        cfg[key] = value
        # TODO


def get_meeting_times(bounds, spacing):
    """
    Generates a list of meeting times given time slot bounds
    and meeting spacings

    PARAMETERS
    -------------
    bounds : list of lists of ints
        [[9,12],[13,17]] or [[900,1200],[1300,1700]]
        = meeting slots 9-12 and 13-17
    spacing : int
        meeting spacing in minutes

    """
    bounds = np.asarray(bounds)

    # check that bounds consist of nested arrays
    if not all(isinstance(b, np.ndarray) for b in bounds):
        raise ValueError("Ill-defined bounds! Bounds must be a list of lists!")

    # check that bounds are integers
    # if not all(isinstance(x, (np.int32, int)) for b in bounds for x in b):
    #     raise ValueError(
    #         "Ill-defined bounds! Bounds must be defined by integers!"
    #     )

    if not all(i < j for i, j in zip(bounds.flatten(), bounds.flatten()[1:])):
        raise ValueError(
            "Ill-defined bounds! Bounds must be strictly increasing!"
        )

    # check that bounds notation is consistent
    item_lengths = np.asarray([len(str(x)) for b in bounds for x in b])
    if np.all(item_lengths <= 2):
        print("Nice! Bounds are defined in full hours!")
        bounds = bounds.astype(np.float64)
    elif np.all((item_lengths > 2) & (item_lengths <= 4)):
        print("Nice! Bounds are defined in hours and minutes!")
        bounds = bounds // 100 + bounds % 100 / 60
    else:
        raise ValueError(
            "Ill-defined bounds! Bounds must denote full hours [[9,12]] or hours and minutes [[900,1200]] *and* be consistent!"
        )

    # check that spacing is an integer
    if not isinstance(spacing, int):
        raise ValueError("Spacing must be an integer!")

    import datetime as dt

    times_list = []
    for b in bounds:
        b0 = dt.timedelta(hours=b[0])
        b1 = dt.timedelta(hours=b[1])
        bx = b0
        while bx < b1:
            times_list.append(bx)
            bx += dt.timedelta(minutes=spacing)

    return times_list
